import numpy as np
import heapq


# Class for computing a(i, j) and b(i, j)
class Oracle:
    def __init__(self, R):
        self.m1 = np.cumsum(R, axis=1)
        self.m2 = np.cumsum(R**2, axis=1)
    
    def inner_centroid(self, i, j):
        a = self.m1[:, j].copy()
        if i > 0:
            a = a - self.m1[:, i - 1]
        a = a / (j - i + 1)
        return a
        
    def inner_cost(self, i, j):
        c = self.m2[:, j].sum()
        if i > 0:
            c -= self.m2[:, i - 1].sum()
        b = self.inner_centroid(i, j)**2
        return c - (j - i + 1)*b.sum()

    def outer_centroid(self, i, j):
        m = self.m1.shape[1]
        a = self.m1[:, j]
        if i > 0:
            a = a - self.m1[:, i - 1]
        a = self.m1[:, -1] - a
        if m - j + i - 1 > 0:
            a = a / (m - j + i - 1)
        return a

    def outer_cost(self, i, j):
        m = self.m1.shape[1]
        c = self.m2[:, j].sum()
        if i > 0:
            c -= self.m2[:, i - 1].sum()
        c = self.m2[:, -1].sum() - c
        b = self.outer_centroid(i, j)**2
        return c - (m - j + i - 1)*b.sum()
    
    def cost(self, i, j):
        return self.inner_cost(i, j) + self.outer_cost(i, j)


# GExact algorithm
def greedy_exact(X, k):
    n, m = X.shape
    S = np.zeros((k, m))
    S[0, :] = 1
    W = np.zeros((n, k))
    R = X.copy()
    prevcost = np.inf
    bestcost = np.sum(R**2)
    rounds = 0
    while bestcost < prevcost:
        prevcost = bestcost
        for r in range(1, k):
            R += np.outer(W[:, 0], S[0, :])
            R += np.outer(W[:, r], S[r, :])
            o = Oracle(R)
            best = None
            bestcost = np.inf
            for i in range(m):
                for j in range(i, m):
                    cand = o.cost(i, j)
                    if cand < bestcost:
                        bestcost = cand
                        best = (i, j)
            i,  j = best
            W[:, 0] = o.outer_centroid(i, j)
            W[:, r] = o.inner_centroid(i, j) - W[:, 0]
            S[r, :] = 0
            S[r, i:(j + 1)] = 1
            R -= np.outer(W[:, 0], S[0, :])
            R -= np.outer(W[:, r], S[r, :])
        rounds += 1
    return W, S, bestcost, rounds


def maxseg(o, m):
    i = 1
    j = 1
    cost = np.inf
    while i < m - 1 or j < m - 1:
        a = o.inner_cost(i, j)
        b = o.outer_cost(i, j)
        cost = min(cost, max(a, b))
        if i == j or (j < m - 1 and a < b):
            j += 1
        else:
            i += 1
    return cost

def est(o, sigma, delta, m):
    cost = np.inf
    best = None
    while sigma > 0:
        i = 1
        j = 1
        besta = 0
        while i < m - 1 or j < m - 1:
            a = o.inner_cost(i, j)
            b = o.outer_cost(i, j)
            if a <= sigma:
                besta = max(besta, a)
            if a + b < cost:
                cost = a + b
                best = i, j
            if i == j or (j < m - 1 and a < sigma):
                j += 1
            else:
                i += 1
        sigma = besta - delta

    return best

# GEst algorithm
def greedy_est(X, k, eps):
    n, m = X.shape
    S = np.zeros((k, m))
    S[0, :] = 1
    W = np.zeros((n, k))
    R = X.copy()
    prevcost = np.inf
    bestcost = np.sum(R**2)
    rounds = 0
    while bestcost < prevcost:
        prevcost = bestcost
        for r in range(1, k):
            R += np.outer(W[:, 0], S[0, :])
            R += np.outer(W[:, r], S[r, :])
            o = Oracle(R)

            tau = maxseg(o, m)
            i,  j = est(o, 2*tau, eps*tau, m)

            W[:, 0] = o.outer_centroid(i, j)
            W[:, r] = o.inner_centroid(i, j) - W[:, 0]
            S[r, :] = 0
            S[r, i:(j + 1)] = 1
            R -= np.outer(W[:, 0], S[0, :])
            R -= np.outer(W[:, r], S[r, :])

        for r in range(1, k):
            R += np.outer(W[:, r], S[r, :])
            i, j = hill(R, W[:, r])
            S[r, :] = 0
            S[r, i:(j + 1)] = 1

            R -= np.outer(W[:, r], S[r, :])

        bestcost = np.sum(R**2)
        rounds += 1
    return W, S, bestcost, rounds

def extract_ranges(S):
    k, m = S.shape
    L = []
    for r in range(1, k):
        ind = np.where(S[r, :])[0]
        if len(ind) > 0:
            L.append((ind[0], ind[-1]))
    return L


# Used to handle the pathological case when S is not full rank
def complete_S(S):
    k, m = S.shape
    L = extract_ranges(S)
    S[1:, :] = 0
    heapq.heapify(L)
    R = [(0, m - 1)]

    cur = (-1, -1)
    consttest = 0
    while len(L) > 0:
        t = heapq.heappop(L)
        if t[0] == cur[0]:
            if t[1] > cur[1]:
                heapq.heappush(L, (cur[1] + 1, t[1]))
        elif t[0] == consttest:
            cur = t
            consttest = t[1] + 1
            if t[1] < m - 1:
                S[len(R),t[0]:(t[1] + 1)] = 1
                R.append(t)
        else:
            cur = t
            S[len(R),t[0]:(t[1] + 1)] = 1
            R.append(t)
    i = 0
    j = 0
    ind = len(R)

    while ind < k and j < m:
        if i == len(R) or R[i][0] > j:
            S[ind, j] = 1
            j += 1
            ind += 1
        elif R[i][0] < j:
            i += 1
        elif R[i][0] == j:
            j += 1
        
            
    return S


def hill(R, w):
    n, m = R.shape
    q = np.zeros((3, m))
    start = np.zeros(m, dtype=int)
    end = np.zeros(m, dtype=int)
    q[0, :] = np.cumsum(np.sum(R**2, axis=0))

    for i in range(m):
        s = R[:, i] - w
        q[1, i] = np.sum(s**2)
        start[i] = i
        if i > 0:
            if q[1, i - 1] < q[0, i - 1]:
                start[i] = start[i - 1]
                q[1, i] += q[1, i - 1]
            else:
                q[1, i] += q[0, i - 1]

    q[2, 0] = np.inf
    for i in range(1, m):
        q[2, i] = np.sum(R[:, i]**2)
        end[i] = i
        if i > 0:
            if q[2, i - 1] < q[1, i - 1]:
                end[i] = end[i - 1]
                q[2, i] += q[2, i - 1]
            else:
                q[2, i] += q[1, i - 1]

    b = np.argmin(q[:, -1])

    if b == 0:
        return (-1, -1)
    if b == 1:
        j = m - 1
        i = start[-1]
    else:
        j = end[-1] - 1
        i = start[j]
    return i, j

# IHill algorithm
def iterative_hill(X, W):
    n, m = X.shape
    k = W.shape[1]
    S = np.zeros((k, m))
    S[0, :] = 1

    prevcost = np.inf
    R = X - W@S
    bestcost = np.sum(R**2)

    rounds = 0
    while bestcost < prevcost:
        prevcost = bestcost


        for r in range(1, k):
            R += np.outer(W[:, r], S[r, :])
            i, j = hill(R, W[:, r])

            S[r, :] = 0
            if i >= 0:
                S[r, i:(j + 1)] = 1

            R -= np.outer(W[:, r], S[r, :])

        S = complete_S(S)
        W = X@S.T@np.linalg.inv(S@S.T + 10e-10*np.eye(k))
        R = X - W@S
        bestcost = np.sum(R**2)
        rounds += 1
    return W, S, bestcost, rounds


def generate_seed(X, k):
    n, m = X.shape
    S = np.zeros((k, m))
    S[0, :] = 1
    for r in range(1, k):
        i = np.random.randint(m)
        j = np.random.randint(m)
        S[r, min(i, j):(1 + max(i, j))] = 1
    return X@S.T@np.linalg.inv(S@S.T + 10e-10*np.eye(k))

def buildlookup(k):
    A = -np.ones((3**k, k), dtype=int)
    for i in range(A.shape[0]):
        j = i
        shift = 1
        ind = 0
        while j > 0:
            if j % 3 > 0:
                A[i, ind] = i - shift
                ind += 1
            shift *= 3
            j //= 3
    return A

def buildlinklookup(k):
    A = np.zeros(3**k, dtype=int)
    mask = np.zeros(3**k, dtype=bool)
    for i in range(A.shape[0]):
        j = i
        shift = 1
        while j > 0:
            if j % 3 == 1:
                A[i] += shift
            elif j % 3 == 2:
                mask[i] = True
            shift *= 2
            j //= 3
    return A, mask

def builddecomp(W):
    n, k = W.shape
    L = np.zeros((n, 2**(k - 1)))
    L[:, 0] = W[:, 0]
    for j in range(k - 1):
        s = 2**j
        for i in range(s, 2*s):
            L[:, i] = L[:, i - s] + W[:, j + 1]
    return L


# Iexact algorithm
def iterative_exact(X, W):
    n, m = X.shape
    k = W.shape[1]
    S = np.zeros((k, m))
    S[0, :] = 1

    prevcost = np.inf
    R = X - W@S
    bestcost = np.sum(R**2)

    tind, mask = buildlinklookup(k - 1)
    sub = buildlookup(k - 1)
    t = tind.shape[0]

    rounds = 0

    while bestcost < prevcost:
        prevcost = bestcost

        L = builddecomp(W)
        q = np.zeros((t, m))
        best = np.zeros((t, m), dtype=int)

        err = np.sum((L - X[:, [0]])**2, axis=0)
        q[:, 0] = err[tind]

        for j in range(1, m):
            for i in range(t):
                q[i, j] = q[i, j - 1]
                best[i, j] = i
                r = 0
                while r < k - 1 and sub[i, r] >= 0:
                    if q[sub[i, r], j] < q[i, j]:
                        q[i, j] = q[sub[i, r], j]
                        best[i, j] = best[sub[i, r], j]
                    r += 1

            err = np.sum((L - X[:, [j]])**2, axis=0)
            q[:, j] += err[tind]

        c = np.argmin(q[:, -1])
        S[1:,:] = 0
        for j in range(m - 1, -1, -1):
            # tind[c] is number with the bits corresponding to the active row in S
            # convert the index into a row in S
            ind = np.where([int(x) for x in reversed(list('{0:0b}'.format(tind[c])))])[0] + 1
            S[ind, j] = 1
            c = best[c, j]


        S = complete_S(S)
        W = X@S.T@np.linalg.inv(S@S.T)
        R = X - W@S
        bestcost = np.sum(R**2)
        rounds += 1

    return W, S, bestcost, rounds
